(* interface between isac math engine and java:
   java -> sml: strings on stdin
   sml -> java: xml on stdout

   WN071004 The xml still reflects the insecurity during the first 
   implementation phase, how the communication via stdin/out could
   correctly relate multiple sml-calculations and java-calculations.

   Since this insecurity turned out unjustified, the xml can be
   simplified in several ways:
   # omit the CALCID; the relation is done by 
     "@@@@@begin@@@@@\n "^string_of_int uI
   # omit the distinctions APPENDFORMULA, REPLACEFORMULA, ...
   WN071004 these 2 simplifications are begun with CALCMESSAGE

   use"xmlsrc/interface-xml.sml";
   use"interface-xml.sml";
   *)

type iterID = int;
type calcID = int;



(** add and delete users -----------------------------------------------
 FIXXME.8.03 addUser: clear code, because only CalcTrees distinguished**)
fun adduserOK2xml (cI:calcID) (uI:iterID) = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int uI^" \n" ^
	     "<ADDUSER>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <USERID> "^string_of_int uI^" </USERID>\n" ^
	     "</ADDUSER>\n" ^
	     "@@@@@end@@@@@");
fun deluserOK2xml (cI:calcID) (uI:iterID) = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int uI^" \n" ^
	     "<DELUSER>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <USERID> "^string_of_int uI^" </USERID>\n" ^
	     "</DELUSER>\n" ^
	     "@@@@@end@@@@@");
(*---------------------------------------------------------------------*)

fun calctreeOK2xml (*uI:iterID*) (cI:calcID) = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CALCTREE>\n" ^
	     "   <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "</CALCTREE>\n" ^
	     "@@@@@end@@@@@");
fun deconstructcalctreeOK2xml (*uI:userID*) (cI:calcID) = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<DELCALC>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "</DELCALC>\n" ^
	     "@@@@@end@@@@@");

fun iteratorOK2xml (cI:calcID) (p:pos')= 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CALCITERATOR>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     pos'2xml i ("POSITION", p) ^
	     "</CALCITERATOR>\n" ^
	     "@@@@@end@@@@@");
fun iteratorERROR2xml (cI:calcID) = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CALCITERATOR>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <ERROR> pos does not exist </ERROR>\n" ^
	     "</CALCITERATOR>\n" ^
	     "@@@@@end@@@@@");

fun sysERROR2xml (cI:calcID) "" = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<SYSERROR>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <ERROR> in kernel </ERROR>\n" ^
	     "</SYSERROR>\n" ^
	     "@@@@@end@@@@@")
  | sysERROR2xml (cI:calcID) str = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<SYSERROR>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <ERROR> "^str^" </ERROR>\n" ^
	     "</SYSERROR>\n" ^
	     "@@@@@end@@@@@");

fun refformulaOK2xml (cI:calcID) p (Form t) = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<REFFORMULA>\n" ^
             "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <CALCFORMULA>\n"^
	     pos'2xml (2*i) ("POSITION", p) ^
	     "    <FORMULA>" ^
	     term2xml (2*i) t ^ "\n" ^
	     "    </FORMULA>\n" ^
	     "  </CALCFORMULA>\n" ^
	     "</REFFORMULA>\n" ^
	     "@@@@@end@@@@@") 
  | refformulaOK2xml (cI:calcID) p (ModSpec modspec) =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<REFFORMULA>\n" ^
             "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     pos'calchead2xml i (p, modspec)^ 
	     "</REFFORMULA>\n" ^
	     "@@@@@end@@@@@"); 

fun refformulaERROR2xml (cI:calcID) = (*FIXME.WN.29.8.03 unused*)
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<REFFORMULA>\n" ^
	     "   <ERROR> object is not a formula </ERROR>\n" ^
	     "</REFFORMULA>\n" ^
	     "@@@@@end@@@@@");

(* val (cI, tac) = (cI, ta);
   *)
fun gettacticOK2xml (cI:calcID) tac = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<GETTACTIC>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n"^
	     tac2xml i tac^
	     "</GETTACTIC>\n" ^
	     "@@@@@end@@@@@");
fun gettacticERROR2xml (cI:calcID) str = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<GETTACTIC>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <ERROR> "^str^" </ERROR>\n" ^
	     "</GETTACTIC>\n" ^
	     "@@@@@end@@@@@");

fun applicabletacticsOK cI tacs =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<APPLICABLETACTICS>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <TACLIST>\n"^
	     tacs2xml (2*i) tacs^
	     "  </TACLIST>\n" ^
	     "</APPLICABLETACTICS>\n" ^
	     "@@@@@end@@@@@");

fun getasmsOK2xml (cI:calcID) terms = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<GETASSUMPTIONS>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <ASMLIST>\n"^
	     formulae2xml (i+i) terms ^
	     "  </ASMLIST>\n" ^
	     "</GETASSUMPTIONS>\n" ^
	     "@@@@@end@@@@@");
(* getasmsOK2xml 333 [str2term "1+1=2", str2term "1+1+1=3"];
   *)

(*WN0502 @see ME/ctree: type asms: illdesigned, thus not used*)
fun getaccuasmsOK2xml cI asms =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<GETACCUMULATEDASMS>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <ASMLIST>\n"^
	     formulae2xml (i+i) asms^
	     "  </ASMLIST>\n" ^
	     "</GETACCUMULATEDASMS>\n" ^
	     "@@@@@end@@@@@");
(* getaccuasmsOK2xml 333 [(([1],Res), str2term "1+1=2"),
			  (([2],Res), str2term "1+1+1=3")];
   getaccuasmsOK2xml 333 [str2term "1+1=2", str2term "1+1+1=3"];
   *)

fun getintervalOK (cI:calcID) fs = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<GETELEMENTSFROMTO>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <FORMHEADS>\n"^	     
	     posterms2xml (2*i) fs^
	     "  </FORMHEADS>\n" ^	     
	     "</GETELEMENTSFROMTO>\n" ^
	     "@@@@@end@@@@@");


fun matchpbl2xml (cI:calcID) (model_ok, pI, hdl, pbl, pre) =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CONTEXTPBL>\n" ^
	     "  <GUH> " ^ pblID2guh pI ^ " </GUH>\n" ^
	     "  <STATUS> " ^ (if model_ok 
			     then "correct" 
			     else "incorrect") ^ " </STATUS>\n" ^
	     "  <HEAD>\n" ^
	     term2xml i hdl ^ "\n" ^
	     "  </HEAD>\n" ^
	     model2xml i pbl pre ^
	     "</CONTEXTPBL>\n" ^
	     "@@@@@end@@@@@");

fun matchmet2xml (cI:calcID) (model_ok, pI, scr, pbl, pre) =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CONTEXTMET>\n" ^
	     "  <GUH> " ^ metID2guh pI ^ " </GUH>\n" ^
	     "  <STATUS> " ^ (if model_ok 
			     then "correct" 
			     else "incorrect") ^ " </STATUS>\n" ^
	     scr2xml i scr ^
	     model2xml i pbl pre ^
	     "</CONTEXTMET>\n" ^
	     "@@@@@end@@@@@");


fun tryrefineOK2xml (cI:calcID) (ModSpec modspec) =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<TRYREFINE>\n" ^
             "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     modspec2xml i modspec ^
	     "</TRYREFINE>\n" ^
	     "@@@@@end@@@@@"); 

fun appendformulaOK2xml (cI:calcID) (old:pos') (del:pos') (new:pos') =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<APPENDFORMULA>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <CALCCHANGED>\n" ^
	     pos'2xml (2*i) ("UNCHANGED", old) ^ 
	     pos'2xml (2*i) ("DELETED", del) ^ 
	     pos'2xml (2*i) ("GENERATED", new) ^ 
	     "  </CALCCHANGED>\n" ^
	     "</APPENDFORMULA>\n" ^
	     "@@@@@end@@@@@");
(* appendformulaOK2xml 1 ([2],Frm) ([3],Pbl) ([4],Res);
   *)
fun appendformulaERROR2xml (cI:calcID) msg =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CALCMESSAGE> "^ msg ^" </CALCMESSAGE>\n" ^
	     "@@@@@end@@@@@");

fun replaceformulaOK2xml (cI:calcID) (old:pos') (del:pos') (new:pos') =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<REPLACEFORMULA>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <CALCCHANGED>\n" ^
	     pos'2xml (2*i) ("UNCHANGED", old) ^ 
	     pos'2xml (2*i) ("DELETED", del) ^ 
	     pos'2xml (2*i) ("GENERATED", new) ^ 
	     "  </CALCCHANGED>\n" ^
	     "</REPLACEFORMULA>\n" ^
	     "@@@@@end@@@@@");
fun replaceformulaERROR2xml (cI:calcID) msg =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CALCMESSAGE> "^ msg ^" </CALCMESSAGE>\n" ^
	     "@@@@@end@@@@@");

fun message2xml (cI:calcID) e = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<MESSAGE>\n" ^
	     "   <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "   <STRING> "^e^" </STRING>\n" ^
	     "</MESSAGE>\n" ^
	     "@@@@@end@@@@@");

fun setnexttactic2xml (*uI:iterID*) (cI:calcID) e = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<SETNEXTTACTIC>\n" ^
	     "   <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "   <MESSAGE> "^e^" </MESSAGE>\n" ^
	     "</SETNEXTTACTIC>\n" ^
	     "@@@@@end@@@@@");

fun fetchproposedtacticOK2xml (*uI:userID*) (cI:calcID) tac = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<NEXTTAC>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n"^
	     tac2xml i tac^
(*	     ^(strs2xml o (map (tac2xml i))) tacs^*)
	     "</NEXTTAC>\n" ^
	     "@@@@@end@@@@@");
(* fetchproposedtactic2xml 11 22 (Rewrite ("rmult_commute","?m *?n =?n *?m"));
   *)
fun fetchproposedtacticERROR2xml (*uI:userID*) (cI:calcID) e = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<NEXTTAC>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <ERROR> "^ e ^" </ERROR>\n" ^
	     "</NEXTTAC>\n" ^
	     "@@@@@end@@@@@");

(*. UNCHANGED: the pos' of the active formula autocalculate has been applied at
    DELETED:   last pos' of the succesional sequence of formulae prob. deleted
    GENERATED: the pos' of the new active formula
.*)
fun autocalculateOK2xml (cI:calcID) (old:pos') (del:pos') (new:pos') = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<AUTOCALC>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <CALCCHANGED>\n" ^
	     pos'2xml (2*i) ("UNCHANGED", old) ^ 
	     pos'2xml (2*i) ("DELETED", del) ^ 
	     pos'2xml (2*i) ("GENERATED", new) ^ 
	     "  </CALCCHANGED>\n" ^
	     "</AUTOCALC>\n" ^
	     "@@@@@end@@@@@");
(* autocalculate2xml 11 22 (Rewrite ("rmult_commute","?m *?n =?n *?m"));
   *)
fun autocalculateERROR2xml (cI:calcID) e = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CALCMESSAGE> "^ e ^" </CALCMESSAGE>\n" ^
	     "@@@@@end@@@@@");

fun interStepsOK (cI:calcID) (*pos'forms*) (old:pos') (del:pos') (new:pos') =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<INTERSTEPS>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <CALCCHANGED>\n" ^
	     pos'2xml (2*i) ("UNCHANGED", old) ^ 
	     pos'2xml (2*i) ("DELETED", del) ^ 
	     pos'2xml (2*i) ("GENERATED", new) ^ 
	     "  </CALCCHANGED>\n" ^
	     "</INTERSTEPS>\n" ^
	     "@@@@@end@@@@@");
fun interStepsERROR (cI:calcID) e =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "  <CALCMESSAGE> "^ e ^" </CALCMESSAGE>\n" ^
	     "@@@@@end@@@@@");

fun modifycalcheadOK2xml (cI:calcID) (chd as (complete,p_,_,_,_,_):ocalhd) =
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<MODIFYCALCHEAD>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     "  <STATUS> "^(if complete then "complete" 
			    else "incomplete")^ "</STATUS>\n"^
	     modspec2xml i chd^
	     "</MODIFYCALCHEAD>\n" ^
	     "@@@@@end@@@@@");

(* val (cI, contthy) = (cI, (context_thy (pt,pos) tac));
   *)
fun contextthyOK2xml cI contthy = 
    writeln ("@@@@@begin@@@@@\n "^string_of_int cI^" \n" ^
	     "<CONTEXTTHY>\n" ^
	     "  <CALCID> "^string_of_int cI^" </CALCID>\n" ^
	     contthy2xml i contthy ^
	     "</CONTEXTTHY>\n" ^
	     "@@@@@end@@@@@");

(*
fun contextthyNO2xml guh = 
    writeln (datatypes.contextthyNO2xml 0 guh);
*)
